from fastapi import HTTPException
from sqlalchemy import inspect
from sqlmodel import Session
from app.models import Validator
from app.services.properties.complex_validator import ComplexValidator
from app.services.properties.property_validator import PropertyValidator
from app.utils import type_casting_utils, enums
from app.utils.admin_utils import admin
from typing import Callable, Dict, Any, List, Optional, Tuple, Type
from app.utils.registrar_utils import get_validation_prefix


class PropertyService:
    def __init__(self, property_records_map: Dict[str, Any], db: Session, entity: str):
        self.property_records_map = property_records_map
        self.institution_synonym_dict = self._load_institution_synonym_dict()
        self.entity = entity
        self.validators = self._load_validators(db, entity)

    def _load_validators(self, db, entity: str) -> List[str]:
        results = db.query(Validator.expression).filter(Validator.entity_type == entity).all()
        return [expr for (expr,) in results]

    def _load_institution_synonym_dict(self) -> Dict[str, Any]:
        return {
            "compound_details": "corporate_compound_id",
            "batch_details": "corporate_batch_id",
        }

    def get_property_info(self, prop_name: str, entity_type: enums.EntityType) -> Dict[str, Any]:
        props = self.property_records_map.get(prop_name)
        if not props:
            raise HTTPException(status_code=400, detail=f"Unknown property: {prop_name}")

        matching_prop = next((p for p in props if getattr(p, "entity_type", None) == entity_type), None)

        if not matching_prop:
            existing_types = [getattr(p, "entity_type", None) for p in props]
            raise HTTPException(
                status_code=400,
                detail=f"Property '{prop_name}' has entity_type(s) {existing_types}, expected '{entity_type.value}'",
            )

        value_type = getattr(matching_prop, "value_type", None)
        if (
            value_type not in type_casting_utils.value_type_to_field
            or value_type not in type_casting_utils.value_type_cast_map
        ):
            raise HTTPException(status_code=400, detail=f"Unsupported or unknown value type for property: {prop_name}")

        return {
            "property": matching_prop,
            "value_type": value_type,
            "field_name": type_casting_utils.value_type_to_field[value_type],
            "cast_fn": type_casting_utils.value_type_cast_map[value_type],
        }

    # TODO: Design a more robust and efficient approach for handling updates to compound details
    def build_details_records(
        self,
        model: Type,
        properties: Dict[str, Any],
        entity_ids: Dict[str, Any],
        entity_type: enums.EntityType,
        include_user_fields: bool = True,
        update_checker: Optional[Callable[[str, int, Any], Optional[Dict[str, Any]]]] = None,
        additional_details: Optional[Dict[str, Any]] = None,
    ) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
        records_to_insert = []
        records_to_validate = {}

        for prop_name, value in properties.items():
            prop_info = self.get_property_info(prop_name, entity_type)
            prop = prop_info["property"]
            value_type = prop_info["value_type"]
            cast_fn = prop_info["cast_fn"]
            field_name = prop_info["field_name"]
            prop_id = getattr(prop, "id")

            if PropertyValidator.validate_nullable(value, prop):
                continue

            value_qualifier, value = self.extract_qualifiers(value_type, value)

            PropertyValidator.validate_value(value, prop)

            try:
                casted_value = cast_fn(value)
            except Exception as e:
                raise HTTPException(status_code=400, detail=f"Error casting value for property {prop_name}: {e}")

            detail = {
                **entity_ids,
                "property_id": prop_id,
                field_name: casted_value,
            }

            records_to_validate.update({prop_name: casted_value})

            # TODO: Refactor to generically handle all value_* fields without hardcoding model-specific attributes
            mapper = inspect(model)
            value_columns = [
                col.key for col in mapper.columns if col.key.startswith("value") and col.key not in field_name
            ]

            for col_name in value_columns:
                if col_name == "value_qualifier":
                    detail[col_name] = value_qualifier
                    continue

                column = mapper.columns[col_name]
                default = None
                if column.default is not None:
                    if not callable(column.default.arg):
                        default = column.default.arg
                detail[col_name] = default

            if include_user_fields:
                detail["created_by"] = admin.admin_user_id
                detail["updated_by"] = admin.admin_user_id

            records_to_insert.append(detail)
        records_to_validate = {f"{get_validation_prefix(entity_type)}": records_to_validate}
        if entity_type.value == self.entity and self.validators and records_to_validate:
            records_to_validate = {**records_to_validate, **(additional_details or {})}
            ComplexValidator.validate_record(records_to_validate, self.validators)

        return records_to_insert, records_to_validate

    def extract_qualifiers(self, value_type: str, value: Any):
        #  Detect and parse value qualifiers
        value_qualifier = 0
        if value_type in ("double", "int") and isinstance(value, str):
            qualifiers = {
                "<": enums.ValueQualifier.LESS_THAN,
                ">": enums.ValueQualifier.GREATER_THAN,
                "=": enums.ValueQualifier.EQUALS,
            }

            first_char = value[0] if value else ""
            if first_char in qualifiers:
                value_qualifier = qualifiers[first_char]
                value = value[1:].strip()
        return value_qualifier, value



from typing import List, Dict, Any
from psycopg2.extensions import adapt
from sqlmodel import SQLModel
from app.utils import enums

column_types = {
    "value_datetime": "timestamptz",
    "value_num": "double precision",
    "value_uuid": "uuid",
    "value_string": "text",
    "value_bool": "boolean",
    "value_qualifier": "smallint",
}


def values_sql(data: List[Dict[str, Any]], columns: List[str]) -> str:
    def escape_val(val, col_name):
        if val is None:
            sql_type = column_types.get(col_name)
            if sql_type:
                return f"NULL::{sql_type}"
            return "NULL"
        return adapt(val).getquoted().decode()

    rows = []
    for row in data:
        values = [escape_val(row.get(col), col) for col in columns]
        rows.append(f"({', '.join(values)})")
    return ",\n".join(rows)


def prepare_sql_parts(records: List[Dict[str, Any]]):
    cols = list(records[0].keys())
    key, *cols_without_key = cols
    values_to_sql = values_sql(records, cols)
    return cols_without_key, values_to_sql


def generate_sql(*sql_parts: str, terminate_with_select: bool = True) -> str:
    filtered_parts = [part.strip() for part in sql_parts if part and part.strip()]
    if not filtered_parts:
        return ""
    combined_sql = ",\n".join(filtered_parts)
    combined_sql += "\nSELECT 1;" if terminate_with_select else ";"
    return combined_sql


def chunked(lst, size):
    for i in range(0, len(lst), size):
        yield lst[i : i + size]


def normalize_type(type_name: str) -> str:
    t = type_name.lower()

    for vt in enums.ValueType:
        if vt.value in t:
            return vt.value
    return enums.ValueType.string.value


def get_table_fields(table_name: str) -> list[dict[str, Any]] | None:
    mapping = {
        "batches": enums.EntityType.BATCH,
        "compounds": enums.EntityType.COMPOUND,
        "assays": enums.EntityType.ASSAY,
        "assay_runs": enums.EntityType.ASSAY_RUN,
        "assay_results": enums.EntityType.ASSAY_RESULT,
    }

    table = SQLModel.metadata.tables.get(f"moltrack.{table_name}")
    if table is None:
        return None

    molecule_columns = {"canonical_smiles", "original_molfile"}

    fields = []
    for col in table.columns:
        fields.append(
            {
                "entity_type": mapping.get(table_name),
                "name": col.name,
                "value_type": normalize_type(col.type.__class__.__name__),
                "semantic_type": {"name": "Molecule", "description": ""} if col.name in molecule_columns else None,
            }
        )

    return fields


def get_direct_fields():
    return [field for table_name in enums.SearchEntityType for field in get_table_fields(table_name.value)]



import re
from typing import List, Optional, Union


class NumericConstraint:
    # Operators
    NONE = "none"
    EQUALS = "="
    NOT_EQUALS = "!="
    GT = ">"
    GTOE = ">="
    LT = "<"
    LTOE = "<="
    IN = "in"
    NOT_IN = "not in"
    RANGE = "-"
    IS_NULL = "is null"
    IS_NOT_NULL = "is not null"

    # Regex patterns
    _num = r"[-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?"
    numRegex = re.compile(f"^{_num}")
    unaryRegex = re.compile(r"(?:[$]{.*})?\s*(!=|>=|<=|=|>|<|\bnot in\b|\bin\b)")
    rangeRegex = re.compile(rf"({_num})\s?(?:-|\.{2})\s?({_num})")
    inRegex = re.compile(r"\((.*)\)")
    isNullRegex = re.compile(r"(^is null$|^is not null$)")

    unaryOperators = [EQUALS, NOT_EQUALS, GT, GTOE, LT, LTOE, IN, NOT_IN]

    def __init__(
        self, op: str, v1: Optional[float] = None, v2: Optional[float] = None, values: Optional[List[float]] = None
    ):
        self.op = op
        self.v1 = v1
        self.v2 = v2
        self.values = values or []

    @staticmethod
    def parse(query: str) -> Optional["NumericConstraint"]:
        query = query.strip()
        if not query:
            return NumericConstraint(NumericConstraint.NONE)

        # Simple numeric → equals
        try:
            x = float(query)
            return NumericConstraint(NumericConstraint.EQUALS, x)
        except ValueError:
            pass

        # Unary operators (=, !=, >, <, etc.)
        match = NumericConstraint.unaryRegex.search(query)
        if match:
            op = match.group(1)
            rest = query[match.end() :].strip()

            if op in (NumericConstraint.IN, NumericConstraint.NOT_IN):
                match_values = NumericConstraint.inRegex.search(rest)
                if match_values:
                    values = [float(s.strip()) for s in match_values.group(1).split(",")]
                    return NumericConstraint(op, values=values)
            else:
                match_value = NumericConstraint.numRegex.match(rest)
                if match_value:
                    v1 = float(match_value.group(0))
                    return NumericConstraint(op, v1=v1)

        # Range like 5-10
        range_match = NumericConstraint.rangeRegex.search(query)
        if range_match:
            return NumericConstraint(NumericConstraint.RANGE, float(range_match.group(1)), float(range_match.group(2)))

        # Null check
        null_match = NumericConstraint.isNullRegex.match(query.lower())
        if null_match:
            return NumericConstraint(null_match.group(1))

        return None

    def is_satisfied_for(self, x: Optional[Union[int, float]]) -> bool:
        if x is None:
            return self.op in (self.NONE, self.IS_NULL)

        if self.op == self.NONE:
            return False
        if self.op == self.GT:
            return x > self.v1
        if self.op == self.GTOE:
            return x >= self.v1
        if self.op == self.LT:
            return x < self.v1
        if self.op == self.LTOE:
            return x <= self.v1
        if self.op == self.EQUALS:
            return x == self.v1
        if self.op == self.NOT_EQUALS:
            return x != self.v1
        if self.op == self.IN:
            return x in self.values
        if self.op == self.NOT_IN:
            return x not in self.values
        if self.op == self.RANGE:
            return self.v1 <= x <= self.v2
        if self.op == self.IS_NULL:
            return False
        if self.op == self.IS_NOT_NULL:
            return True
        raise ValueError(f"Unknown operation {self.op}")



from fastapi import HTTPException
from sqlalchemy import and_
from sqlalchemy.orm import Session
from sqlalchemy.orm import joinedload
from app.crud.properties import enrich_model
from app import crud, models


def enrich_batch(batch: models.Batch) -> models.BatchResponse:
    batch_resp = enrich_model(batch, models.BatchResponse, "batch_details", "batch_id")
    if batch.compound:
        batch_resp.compound = enrich_model(batch.compound, models.CompoundResponse, "compound_details", "compound_id")
    return batch_resp


def get_batch_by_synonym(db: Session, property_value: str, property_name: str = None, enrich: bool = True):
    if not property_value:
        return None

    filters = [models.Property.semantic_type_id == 1, models.BatchDetail.value_string == property_value]

    if property_name:
        filters.append(models.Property.name == property_name)

    batch = (
        db.query(models.Batch)
        .join(models.Batch.batch_details)
        .join(models.BatchDetail.property)
        .options(joinedload(models.Batch.batch_details).joinedload(models.BatchDetail.property))
        .filter(and_(*filters))
        .first()
    )

    if not batch:
        return None

    return enrich_batch(batch) if enrich else batch


def get_batches(db: Session, skip: int = 0, limit: int = 100):
    batches = db.query(models.Batch).offset(skip).limit(limit).all()
    return [enrich_batch(batch) for batch in batches]


def get_batches_by_compound(db: Session, compound_id: int, skip: int = 0, limit: int = 100):
    return db.query(models.Batch).filter(models.Batch.compound_id == compound_id).offset(skip).limit(limit).all()


def delete_batch_by_synonym(
    db: Session,
    property_value: str,
    property_name: str,
):
    db_batch = get_batch_by_synonym(db, property_value, property_name, False)
    if db_batch is None:
        raise HTTPException(status_code=404, detail="Batch not found")

    assay_results = crud.get_all_assay_results_for_batch(db, db_batch.id)
    if assay_results:
        raise HTTPException(status_code=400, detail="Batch has dependent assay results")

    db.delete(db_batch)
    db.commit()
    return db_batch



from datetime import datetime
from typing import Dict, Mapping, List, Any
from cel import evaluate, Context
import re

from app.utils import enums


class ComplexValidationError(ValueError):
    """Raised when a record does not satisfy a validation expression."""


class ComplexValidator:
    """
    Validates entire records (multi-field objects) using CEL expressions.
    Rules can use placeholders like ${field}, ${field}.length, is null, etc.
    """

    @classmethod
    def validate_record(cls, record: Dict[str, Any], rules: List[str] = None, validate_rule=False) -> None:
        """
        Validate a record against a list of CEL rules.
        Raises RecordValidationError if any rule fails.
        """
        if not rules:
            return
        safe_ctx = cls._sanitize_context(record)
        ctx = cls._build_context(safe_ctx)

        for raw_expr in rules:
            evaluate_rule = True
            variables = cls._extract_variables(raw_expr)
            for var in variables:
                var_parts = var.replace(" ", "_").split(".")
                details_table, property_name = var_parts[0], var_parts[1]
                if details_table not in safe_ctx.keys() or property_name not in safe_ctx[details_table]:
                    evaluate_rule = False
                    break
            if not evaluate_rule:
                continue
            expr = cls._preprocess(raw_expr)
            try:
                result = evaluate(expr, ctx, mode="strict")
            except Exception as e:
                raise ComplexValidationError(
                    f"Error while evaluating rule '{raw_expr}' (translated to '{expr}'): {e}"
                ) from e

            if not validate_rule and not bool(result):
                raise ComplexValidationError(f"Record does not satisfy rule: {raw_expr}")

    @classmethod
    def validate_rule(cls, expr: str, properties: Dict[str, str], entity_type: enums.EntityType) -> bool:
        """
        Validate a single CEL rule against a dictionary of properties.
        Creates a mock context where each property is set to a non-null value.
        Returns True if the rule compiles, False otherwise.
        """

        variables = cls._extract_variables(expr)
        for i, var in enumerate(variables):
            standardized_var = var
            if "." not in var:
                standardized_var = f"{entity_type.value.lower()}_details.{var}"
                expr = expr.replace(f"${{{var}}}", f"${{{standardized_var}}}")
            if standardized_var not in properties.keys():
                raise ComplexValidationError(
                    f"Unknown property '{var}' (transformed to {standardized_var}) in rule: {expr}"
                )
            variables[i] = standardized_var

        mock_record = cls._create_mock_record(variables, properties)

        cls.validate_record(mock_record, [expr], validate_rule=True)

        return expr

    # ------------------------
    # Internal helpers
    # ------------------------
    @staticmethod
    def _sanitize_context(record: Dict[str, Any]) -> Mapping[str, Any]:
        """Convert record values into JSON-serializable primitives."""

        safe_ctx = {}
        for scope in record.keys():
            scope_details = {}
            for k, v in record[scope].items():
                k = k.replace(" ", "_")
                if isinstance(v, (str, int, float, bool)) or v is None:
                    scope_details[k] = v
                else:
                    scope_details[k] = str(v)
            safe_ctx[scope] = scope_details
        return safe_ctx

    @staticmethod
    def _build_context(record: Dict[str, Any]) -> Context:
        """
        Build a CEL Context with custom functions.
        """

        ctx = Context()

        for key, value in record.items():
            ctx.add_variable(key, value)

        # size(value) → length of string or list
        ctx.add_function("size", lambda x: len(x) if x is not None else 0)

        # matches(value, pattern) → regex match
        ctx.add_function("matches", lambda val, pat: re.match(pat, val) is not None)

        # today() → current date as ISO string
        ctx.add_function("today", lambda: datetime.today().date().isoformat())

        # date(str) → convert string to date object
        ctx.add_function("date", lambda s: datetime.fromisoformat(s).date())

        return ctx

    @staticmethod
    def _preprocess(expr: str) -> str:
        """Translate DSL-style expressions into CEL syntax."""
        # ${field} → field
        expr = re.sub(r"\$\{([a-zA-Z_][a-zA-Z0-9_]*)\}", r"\1", expr)

        # ${field}.length → size(field)
        expr = re.sub(r"([a-zA-Z_][a-zA-Z0-9_]*)\.length", r"size(\1)", expr)

        # is null → == null
        expr = re.sub(r"([a-zA-Z_][a-zA-Z0-9_]*)\s+is\s+null", r"\1 == null", expr)

        # is not null → != null
        expr = re.sub(r"([a-zA-Z_][a-zA-Z0-9_]*)\s+is\s+not\s+null", r"\1 != null", expr)

        # replaces white spaces inside ${...} with underscores
        expr = re.sub(r"\$\{([^}]+)\}", lambda m: re.sub(r"[^a-zA-Z0-9_.]", "_", m.group(1)), expr)

        return expr.strip()

    @staticmethod
    def _extract_variables(expr: str) -> list[str]:
        """
        Extract all variables from an expression.
        """

        pattern = re.compile(r"\$\{([^}]+)\}")
        vars = pattern.findall(expr)
        vars_cel = {re.sub(r"\$\{([a-zA-Z_][a-zA-Z0-9_]*)\}", r"\1", v) for v in vars}
        return list(vars_cel)

    @staticmethod
    def _create_mock_record(variables, properties) -> dict[str, Any]:
        """
        Create a mock record with sample values for testing.
        """

        mock_values = {
            "string": "mock",
            "int": 2,
            "double": 2.0,
            "bool": True,
            "datetime": datetime.now().isoformat(),
            "uuid": "123e4567-e89b-12d3-a456-426614174000",
        }

        record = {}
        for var in variables:
            parent, child = var.split(".", 1)
            record.setdefault(parent, {})
            value = mock_values.get(properties.get(var, "string"), "mock")
            record[parent][child] = value

        return record



from typing import Dict, Any, Type
from app import models
from app.utils import enums
from sqlalchemy.orm import Session


def get_validation_prefix(entity_type: enums.EntityType) -> str:
    validation_name_map = {
        enums.EntityType.COMPOUND: "compound_details",
        enums.EntityType.BATCH: "batch_details",
        enums.EntityType.ASSAY: "assay_details",
        enums.EntityType.ASSAY_RUN: "assay_run_details",
        enums.EntityType.ASSAY_RESULT: "assay_result_details",
    }
    return validation_name_map.get(entity_type, "")


def get_details_for_entity(
    entity_id: int,
    cache: Dict[int, Dict[str, Any]],
    entity_type: enums.EntityType,
    db: Session,
    detail_model: Type,
    foreign_key_field: str,
) -> Dict[str, Any]:
    if entity_id in cache:
        return cache[entity_id]

    properties = (
        db.query(detail_model, models.Property)
        .join(models.Property, detail_model.property_id == models.Property.id)
        .filter(getattr(detail_model, foreign_key_field) == entity_id)
        .all()
    )

    prop_dict: Dict[str, Any] = {}
    for detail, prop in properties:
        entity_scope = f"{get_validation_prefix(entity_type)}"
        prop_name = f"{prop.name}"

        scope_dict = prop_dict.setdefault(entity_scope, {})
        match prop.value_type:
            case enums.ValueType.string:
                scope_dict[prop_name] = detail.value_string
            case enums.ValueType.int | enums.ValueType.double:
                scope_dict[prop_name] = detail.value_num
            case enums.ValueType.datetime:
                scope_dict[prop_name] = detail.value_datetime
            case enums.ValueType.uuid:
                scope_dict[prop_name] = detail.value_uuid

    cache[entity_id] = prop_dict
    return prop_dict



from app.utils.enums import AggregationNumericOp, AggregationStringOp


class AggregationOperators:
    """Maps aggregation operators to SQL expressions and validation"""

    OPERATORS = {
        AggregationNumericOp.TOTAL_COUNT.value: "COUNT(*)",
        AggregationNumericOp.VALUE_COUNT.value: "COUNT({column})",
        AggregationNumericOp.UNIQUE_COUNT.value: "COUNT(DISTINCT {column})",
        AggregationNumericOp.MISSING_VALUE_COUNT.value: "SUM(({column} IS NULL)::int)",
        AggregationNumericOp.MIN.value: "MIN(CAST({column} AS NUMERIC))",
        AggregationNumericOp.MAX.value: "MAX(CAST({column} AS NUMERIC))",
        AggregationNumericOp.SUM.value: "SUM(CAST({column} AS NUMERIC))",
        AggregationNumericOp.MED.value: "PERCENTILE_CONT(0.5) WITHIN GROUP (ORDER BY CAST({column} AS NUMERIC))",
        AggregationNumericOp.AVG.value: "AVG(CAST({column} AS NUMERIC))",
        AggregationNumericOp.STDEV.value: "STDDEV_SAMP(CAST({column} AS NUMERIC))",
        AggregationNumericOp.VARIANCE.value: "VAR_SAMP(CAST({column} as NUMERIC))",
        AggregationNumericOp.STDDEV_POP.value: "STDDEV_POP(CAST({column} AS NUMERIC))",
        AggregationNumericOp.VAR_POP.value: "VAR_POP(CAST({column} as NUMERIC))",
        AggregationNumericOp.Q1.value: "PERCENTILE_CONT(0.25) WITHIN GROUP (ORDER BY CAST({column} AS NUMERIC))",
        AggregationNumericOp.Q2.value: "PERCENTILE_CONT(0.5) WITHIN GROUP (ORDER BY CAST({column} AS NUMERIC))",
        AggregationNumericOp.Q3.value: "PERCENTILE_CONT(0.75) WITHIN GROUP (ORDER BY CAST({column} AS NUMERIC))",
        AggregationNumericOp.ARRAY_AGG.value: "ARRAY_AGG({column})",
        # Note: AggregationStringOp.FIRST does not return the first value by insertion order.
        # Instead, it returns the lexicographically smallest string, which is usually fine
        # when we just want any value.
        AggregationStringOp.FIRST.value: "MIN({column})",
        AggregationStringOp.CONCAT_ALL.value: "STRING_AGG({column}, ', ')",
        AggregationStringOp.CONCAT_UNIQUE.value: "STRING_AGG(DISTINCT {column}, ', ')",
        AggregationStringOp.MOST_FREQUENT.value: "MODE() WITHIN GROUP (ORDER BY {column}) ",
    }

    @classmethod
    def get_sql_expression(cls, operator: str | None, column: str, condition: str) -> str:
        if not operator:
            return f"MAX({column})"

        try:
            return cls.OPERATORS[operator].format(column=column)
        except KeyError:
            raise ValueError(f"Unsupported operator: {operator}")



# app/crud/__init__.py

from .compounds import (
    get_compound_by_synonym,
    get_compound_by_hash,
    read_compounds,
    update_compound,
    delete_compound
)

from .properties import (
    create_properties,
    get_properties,
    get_entities_by_entity_type,
    get_synonym_id,
    update_property_vocabulary,
)

from .batches import get_batches_by_compound, get_batches, get_batch_by_synonym, delete_batch_by_synonym

from .additions import (
    create_additions,
    get_additions,
    get_addition_by_id,
    update_addition_by_id,
    delete_addition_by_id,
    get_batch_addition_for_addition,
)

from .assay_data import (
    get_assays,
    get_assay,
    get_assay_runs,
    get_assay_run,
    get_all_assay_results_for_batch,
    get_assay_results,
)

from .validators import get_validators_for_entity, delete_validator_by_name, create_validator



import csv
import io
import pytest
import json
from app.utils import enums


def send_registration_request(client, rows):
    csv_buffer = io.StringIO()
    writer = csv.DictWriter(csv_buffer, fieldnames=rows[0].keys())
    writer.writeheader()
    writer.writerows(rows)
    csv_buffer.seek(0)

    files = {"file": ("data.csv", csv_buffer.getvalue(), "text/csv")}
    data = {"error_handling": enums.ErrorHandlingOptions.reject_row.value}

    response = client.post("v1/compounds/", files=files, data=data)
    return response


validated_property_validator = {
    "properties": [
        {
            "name": "val_int_prop_1",
            "description": "val_int_prop_1 description",
            "value_type": "int",
            "property_class": "DECLARED",
            "unit": "",
            "entity_type": "COMPOUND",
            "pattern": "",
            "validators": '[">5", "<15"]',
        }
    ]
}


rows_validators = [
    {"smiles": "Brc1cc(Br)c2ccn(Br)c2c1", "val_int_prop_1": 10},
    {"smiles": "Cc1cc(Br)cc2c1ccn2Br", "val_int_prop_1": 16},
]


@pytest.mark.usefixtures("preload_simple_data")
def test_single_property_validators(client):
    schema_response = client.post("/v1/schema/", json=validated_property_validator)

    schema_response_content = schema_response.content.decode("utf-8")
    schema_response_content = json.loads(schema_response_content)

    assert schema_response_content["status"] == "success"
    assert schema_response_content["property_types"][0]["validators"] is not None

    response = send_registration_request(client, rows_validators)

    assert response.status_code == 200
    assert response.headers["content-disposition"].startswith("attachment;")

    content = response.content.decode("utf-8")
    content = json.loads(content)

    assert len(content) == 2
    assert content[0]["registration_status"] == "success"
    assert content[1]["registration_status"] == "failed"


validated_property_min_max = {
    "properties": [
        {
            "name": "val_int_prop_2",
            "description": "val_int_prop_2 description",
            "value_type": "int",
            "property_class": "DECLARED",
            "unit": "",
            "entity_type": "COMPOUND",
            "pattern": "",
            "min": 5,
            "max": 15,
        }
    ]
}


rows_min_max = [
    {"smiles": "Brc1cc(Br)c2ccn(Br)c2c1", "val_int_prop_2": 10},
    {"smiles": "Cc1cc(Br)cc2c1ccn2Br", "val_int_prop_2": 16},
]


@pytest.mark.usefixtures("preload_simple_data")
def test_single_property_min_max(client):
    schema_response = client.post("/v1/schema/", json=validated_property_min_max)

    schema_response_content = schema_response.content.decode("utf-8")
    schema_response_content = json.loads(schema_response_content)

    assert schema_response_content["status"] == "success"
    assert schema_response_content["property_types"][0]["min"] is not None
    assert schema_response_content["property_types"][0]["max"] is not None

    response = send_registration_request(client, rows_min_max)

    assert response.status_code == 200
    assert response.headers["content-disposition"].startswith("attachment;")

    content = response.content.decode("utf-8")
    content = json.loads(content)

    assert len(content) == 2
    assert content[0]["registration_status"] == "success"
    assert content[1]["registration_status"] == "failed"


validated_property_choices = {
    "properties": [
        {
            "name": "val_string_prop_1",
            "description": "val_string_prop_1 description",
            "value_type": "string",
            "property_class": "DECLARED",
            "unit": "",
            "entity_type": "COMPOUND",
            "pattern": "",
            "choices": '["choice1", "choice2", "choice3"]',
        }
    ]
}


rows_choices = [
    {"smiles": "Brc1cc(Br)c2ccn(Br)c2c1", "val_string_prop_1": "choice2"},
    {"smiles": "Cc1cc(Br)cc2c1ccn2Br", "val_string_prop_1": "choice4"},
]


@pytest.mark.usefixtures("preload_simple_data")
def test_single_property_choices(client):
    schema_response = client.post("/v1/schema/", json=validated_property_choices)

    schema_response_content = schema_response.content.decode("utf-8")
    schema_response_content = json.loads(schema_response_content)

    assert schema_response_content["status"] == "success"
    assert schema_response_content["property_types"][0]["choices"] is not None

    response = send_registration_request(client, rows_choices)

    assert response.status_code == 200
    assert response.headers["content-disposition"].startswith("attachment;")

    content = response.content.decode("utf-8")
    content = json.loads(content)

    assert len(content) == 2
    assert content[0]["registration_status"] == "success"
    assert content[1]["registration_status"] == "failed"


properties = {
    "properties": [
        {
            "name": "val_string_prop_1",
            "description": "val_string_prop_1 description",
            "value_type": "string",
            "property_class": "DECLARED",
            "unit": "",
            "entity_type": "COMPOUND",
            "pattern": "",
        },
        {
            "name": "val_int_prop_1",
            "description": "val_int_prop_1 description",
            "value_type": "int",
            "property_class": "DECLARED",
            "unit": "",
            "entity_type": "COMPOUND",
            "pattern": "",
        },
    ]
}


rows_record_validation = [
    {"smiles": "Brc1cc(Br)c2ccn(Br)c2c1", "val_string_prop_1": "input_1", "val_int_prop_1": 11},
    {"smiles": "Cc1cc(Br)cc2c1ccn2Br", "val_string_prop_1": "inputnotok", "val_int_prop_1": 16},
    {"smiles": "CC(=O)CC(=O)c1cc(Br)cc2c1ccn2Br", "val_string_prop_1": "input_ok", "val_int_prop_1": 2},
]


valid_rule = "matches(${val_string_prop_1}, r'^input_') && ${val_int_prop_1} > 10"


@pytest.mark.usefixtures("preload_simple_data")
def test_validator_registration_and_record_validation_valid(client):
    # Register properties
    schema_response = client.post("/v1/schema/", json=properties)

    schema_response_content = schema_response.content.decode("utf-8")
    schema_response_content = json.loads(schema_response_content)

    assert schema_response_content["status"] == "success"

    # Register validator
    validator_payload = {
        "name": "test_validator_1",
        "entity_type": enums.EntityType.COMPOUND.value,
        "expression": valid_rule,
        "description": "test validator 1 description",
    }
    register_rule_response = client.post("/v1/validators/", data=validator_payload)

    register_rule_response_content = register_rule_response.content.decode("utf-8")
    register_rule_response_content = json.loads(register_rule_response_content)

    assert register_rule_response_content["status"] == "success"
    assert register_rule_response_content["added_validator"] == valid_rule

    response = send_registration_request(client, rows_record_validation)

    assert response.status_code == 200
    assert response.headers["content-disposition"].startswith("attachment;")

    content = response.content.decode("utf-8")
    content = json.loads(content)

    assert len(content) == 3
    assert content[0]["registration_status"] == "success"
    assert content[1]["registration_status"] == "failed"
    assert content[2]["registration_status"] == "failed"


invalid_rule = "matches(${val_string_prop_1}, r'^input_') & ${val_int_prop_1} > 10"


@pytest.mark.usefixtures("preload_simple_data")
def test_validator_registration_invalid_property(client):
    # Register validator
    validator_payload = {
        "name": "test_validator_1",
        "entity_type": enums.EntityType.COMPOUND.value,
        "expression": invalid_rule,
        "description": "test validator 1 description",
    }
    register_rule_response = client.post("/v1/validators/", data=validator_payload)

    register_rule_response_content = register_rule_response.content.decode("utf-8")
    register_rule_response_content = json.loads(register_rule_response_content)

    assert register_rule_response_content["detail"]["status"] == "failed"
    assert register_rule_response_content["detail"]["message"].startswith("Error adding validators:")


@pytest.mark.usefixtures("preload_simple_data")
def test_validator_registration_invalid_rule(client):
    # Register properties
    schema_response = client.post("/v1/schema/", json=properties)

    schema_response_content = schema_response.content.decode("utf-8")
    schema_response_content = json.loads(schema_response_content)

    assert schema_response_content["status"] == "success"

    # Register validator
    validator_payload = {
        "name": "test_validator_1",
        "entity_type": enums.EntityType.COMPOUND.value,
        "expression": invalid_rule,
        "description": "test validator 1 description",
    }
    register_rule_response = client.post("/v1/validators/", data=validator_payload)

    register_rule_response_content = register_rule_response.content.decode("utf-8")
    register_rule_response_content = json.loads(register_rule_response_content)

    assert register_rule_response_content["detail"]["status"] == "failed"
    assert register_rule_response_content["detail"]["message"].startswith("Error adding validators:")



import json
from typing import Any, Dict, List, NamedTuple, Optional, Union, Literal, get_args
from pydantic import ConfigDict, field_validator, model_validator, validator
from sqlalchemy import Column, DateTime, Enum, CheckConstraint
from sqlmodel import SQLModel, Field, Relationship
from sqlalchemy.sql import func
from app.utils import enums
import os
from datetime import datetime, timezone
import uuid
# import crud
# # Handle both package imports and direct execution
# try:
#     # When imported as a package (for tests)
#     from .database import Base
# except ImportError:
#     # When run directly
#     from database import Base

# Get the schema name from environment variable or use default
DB_SCHEMA = os.environ.get("DB_SCHEMA", "moltrack")


class UserBase(SQLModel):
    email: str = Field(nullable=False)
    first_name: str = Field(nullable=False)
    last_name: str = Field(nullable=False)
    is_active: bool = Field(default=False, nullable=False)
    is_service_account: bool = Field(default=False, nullable=False)


class User(UserBase, table=True):
    __tablename__ = "users"
    __table_args__ = {"schema": DB_SCHEMA}

    id: uuid.UUID = Field(primary_key=True, nullable=False, default_factory=uuid.uuid4)
    has_password: bool = Field(nullable=False)
    created_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now(), nullable=False))
    updated_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now(), onupdate=func.now(), nullable=False)
    )
    created_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    updated_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)


class Settings(SQLModel, table=True):
    __tablename__ = "settings"
    __table_args__ = {"schema": DB_SCHEMA}

    id: int = Field(primary_key=True, nullable=False)
    name: str = Field(nullable=False)
    value: str = Field(nullable=False)
    description: str = Field(nullable=False)


class CompoundQueryParams(SQLModel):
    substructure: Optional[str] = None
    skip: int = 0
    limit: int = 100


class CompoundBase(SQLModel):
    original_molfile: Optional[str] = Field(default=None)  # as sketched by the chemist
    is_archived: Optional[bool] = Field(default=False)


class CompoundCreate(CompoundBase):
    smiles: str


class CompoundResponseBase(CompoundBase):
    id: int = Field(primary_key=True, index=True)
    molregno: int = Field(nullable=False)
    canonical_smiles: str = Field(nullable=False)
    inchi: str = Field(nullable=False)
    inchikey: str = Field(nullable=False, unique=True)
    created_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now()))
    updated_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now(), onupdate=func.now())
    )


class CompoundResponse(CompoundResponseBase):
    properties: Optional[List["PropertyWithValue"]] = []


class CompoundDetailBase(SQLModel):
    compound_id: int = Field(foreign_key="moltrack.compounds.id", nullable=False)
    property_id: int = Field(foreign_key="moltrack.properties.id", nullable=False)


class CompoundDetailCreate(CompoundDetailBase):
    value: Any


class CompoundDetail(CompoundDetailBase, table=True):
    __tablename__ = "compound_details"
    __table_args__ = {"schema": DB_SCHEMA}

    id: int = Field(primary_key=True, index=True)
    created_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now(), nullable=False))
    updated_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now(), onupdate=func.now(), nullable=False)
    )
    created_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    updated_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)

    value_datetime: Optional[datetime] = Field(sa_column=Column(DateTime(timezone=True)))
    value_uuid: Optional[uuid.UUID] = Field(default_factory=uuid.uuid4)
    value_num: Optional[float]
    value_string: Optional[str]
    value_qualifier: Optional[int]

    compound: "Compound" = Relationship(back_populates="compound_details")
    property: "Property" = Relationship(back_populates="compound_details")


class Compound(CompoundResponseBase, table=True):
    __tablename__ = "compounds"
    __table_args__ = {"schema": DB_SCHEMA}

    formula: str = Field(nullable=False)
    hash_mol: str = Field(nullable=False)
    hash_tautomer: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    hash_canonical_smiles: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    hash_no_stereo_smiles: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    hash_no_stereo_tautomer: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    created_by: uuid.UUID = Field(foreign_key=f"{DB_SCHEMA}.users.id", nullable=False, default_factory=uuid.uuid4)
    updated_by: uuid.UUID = Field(foreign_key=f"{DB_SCHEMA}.users.id", nullable=False, default_factory=uuid.uuid4)
    deleted_at: Optional[datetime] = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now()))
    deleted_by: Optional[uuid.UUID] = Field(foreign_key=f"{DB_SCHEMA}.users.id", default=None)

    batches: List["Batch"] = Relationship(back_populates="compound")
    compound_details: List["CompoundDetail"] = Relationship(
        back_populates="compound", sa_relationship_kwargs={"cascade": "all, delete-orphan"}
    )
    properties: List["Property"] = Relationship(
        back_populates="compounds", link_model=CompoundDetail, sa_relationship_kwargs={"viewonly": True}
    )


class CompoundDetailUpdate(SQLModel):
    property_id: int
    value: Any


class CompoundUpdate(SQLModel):
    original_molfile: Optional[str] = None
    is_archived: Optional[bool] = None
    canonical_smiles: Optional[str] = None
    properties: Optional[List[CompoundDetailUpdate]] = None


class BatchDetailBase(SQLModel):
    batch_id: int = Field(foreign_key=f"{DB_SCHEMA}.batches.id", nullable=False)
    property_id: int = Field(foreign_key=f"{DB_SCHEMA}.properties.id", nullable=False)

    value_qualifier: Optional[int] = Field(default=0, nullable=False)  # 0 for "=", 1 for "<", 2 for ">"
    value_datetime: Optional[datetime] = Field(default=None, sa_column=Column(DateTime(timezone=True)))
    value_uuid: Optional[uuid.UUID] = Field(default_factory=uuid.uuid4)
    value_num: Optional[float] = Field(default=None)
    value_string: Optional[str] = Field(default=None)


class BatchDetailResponse(BatchDetailBase):
    id: int = Field(primary_key=True, index=True)


class BatchDetail(BatchDetailResponse, table=True):
    __tablename__ = "batch_details"
    __table_args__ = {"schema": DB_SCHEMA}

    created_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now(), nullable=False))
    updated_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now(), nullable=False))
    created_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    updated_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)

    batch: "Batch" = Relationship(back_populates="batch_details")
    property: "Property" = Relationship(back_populates="batch_details")


class BatchBase(SQLModel):
    notes: Optional[str] = Field(default=None)
    compound_id: int = Field(foreign_key=f"{DB_SCHEMA}.compounds.id", nullable=False)


class BatchResponseBase(BatchBase):
    id: int = Field(primary_key=True, index=True)
    created_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now(), nullable=False))


class BatchResponse(BatchResponseBase):
    batch_additions: List["BatchAddition"] = []
    properties: Optional[List["PropertyWithValue"]] = []
    compound: Optional["CompoundResponse"] = None


class Batch(BatchResponseBase, table=True):
    __tablename__ = "batches"
    __table_args__ = {"schema": DB_SCHEMA}

    updated_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now(), nullable=False))
    created_by: uuid.UUID = Field(foreign_key=f"{DB_SCHEMA}.users.id", nullable=False, default_factory=uuid.uuid4)
    updated_by: uuid.UUID = Field(foreign_key=f"{DB_SCHEMA}.users.id", nullable=False, default_factory=uuid.uuid4)
    batch_regno: int = Field(nullable=False)

    compound: "Compound" = Relationship(back_populates="batches")
    assay_results: List["AssayResult"] = Relationship(back_populates="batch")
    batch_details: List["BatchDetail"] = Relationship(
        back_populates="batch", sa_relationship_kwargs={"cascade": "all, delete-orphan"}
    )
    batch_additions: List["BatchAddition"] = Relationship(
        back_populates="batch", sa_relationship_kwargs={"cascade": "all, delete-orphan"}
    )
    properties: List["Property"] = Relationship(
        back_populates="batches", link_model=BatchDetail, sa_relationship_kwargs={"viewonly": True}
    )


class SemanticTypeBase(SQLModel):
    name: str = Field(nullable=False)
    description: Optional[str] = Field(default=None, nullable=True)


class SemanticType(SemanticTypeBase, table=True):
    __tablename__ = "semantic_types"
    __table_args__ = {"schema": DB_SCHEMA}

    id: int = Field(primary_key=True, nullable=False)
    properties: List["Property"] = Relationship(back_populates="semantic_type")


class PropertyBase(SQLModel):
    name: str = Field(nullable=False)
    description: Optional[str] = Field(default=None)
    value_type: enums.ValueType = Field(sa_column=Column(Enum(enums.ValueType), nullable=False))
    property_class: enums.PropertyClass = Field(sa_column=Column(Enum(enums.PropertyClass), nullable=False))
    unit: Optional[str] = Field(default=None)
    entity_type: enums.EntityType = Field(sa_column=Column(Enum(enums.EntityType), nullable=False))
    semantic_type_id: Optional[int] = Field(foreign_key=f"{DB_SCHEMA}.semantic_types.id", nullable=True, default=None)
    pattern: Optional[str] = Field(default=None)
    min: Optional[float] = Field(default=None)
    max: Optional[float] = Field(default=None)
    choices: Optional[str] = Field(default=None)
    validators: Optional[str] = Field(default=None)
    friendly_name: Optional[str] = Field(default=None)
    nullable: bool = Field(default=True)
    created_at: Optional[datetime] = Field(
        default_factory=lambda: datetime.now(timezone.utc),
        sa_column=Column(DateTime(timezone=True), server_default=func.now()),
    )

    @validator("choices", pre=True)
    def serialize_choices(cls, v):
        if isinstance(v, list):
            return json.dumps(v)
        return v


class PropertyInput(PropertyBase):
    semantic_type_name: Optional[str] = Field(default=None)


class PropertyRetrieve(PropertyBase):
    semantic_type: Optional[SemanticTypeBase] = None


class PropertyWithValue(PropertyBase):
    value_datetime: Optional[datetime] = None
    value_uuid: Optional[uuid.UUID] = None
    value_num: Optional[float] = None
    value_string: Optional[str] = None
    value_qualifier: Optional[str] = None


class SynonymTypeBase(PropertyInput):
    """A specialized type of PropertyInput for synonyms with fixed constraints"""

    value_type: enums.ValueType = Field(default=enums.ValueType.string)
    property_class: enums.PropertyClass = Field(default=enums.PropertyClass.DECLARED)
    unit: Optional[str] = Field(default="")
    semantic_type_id: Optional[int] = Field(default=1)

    model_config = ConfigDict(populate_by_name=True)

    @field_validator("value_type")
    def validate_value_type(cls, v):
        if v != enums.ValueType.string:
            raise ValueError("SynonymType must have value_type of string")
        return v

    @field_validator("property_class")
    def validate_property_class(cls, v):
        if v != enums.PropertyClass.DECLARED:
            raise ValueError("SynonymType must have property_class of DECLARED")
        return v

    @field_validator("unit")
    def validate_unit(cls, v):
        if v not in (None, ""):
            raise ValueError("SynonymType must have empty unit")
        return v


class SynonymBatchRequest(SQLModel):
    property_values: List[str]
    property_name: Optional[str] = None


class PropertyResponse(PropertyBase):
    id: int = Field(primary_key=True, index=True)


class AssayProperty(SQLModel, table=True):
    __tablename__ = "assay_properties"
    __table_args__ = {"schema": DB_SCHEMA}

    assay_id: int = Field(foreign_key=f"{DB_SCHEMA}.assays.id", primary_key=True)
    property_id: int = Field(foreign_key=f"{DB_SCHEMA}.properties.id", primary_key=True)
    required: bool = Field(default=False)

    assay: "Assay" = Relationship(back_populates="property_requirements")
    property: "Property" = Relationship()


class AssayBase(SQLModel):
    name: str = Field(nullable=False)
    description: Optional[str] = Field(default=None)


class AssayCreate(AssayBase):
    property_ids: List[int] = []  # List of property IDs to associate with this assay type
    property_requirements: List[Dict[str, Any]] = []  # List of property requirements
    property_details: List[Dict[str, Any]] = []  # List of property metadata


class AssayResponseBase(AssayBase):
    id: int = Field(primary_key=True, index=True)
    created_at: datetime = Field(sa_column=Column(DateTime, server_default=func.now()))
    updated_at: datetime = Field(sa_column=Column(DateTime, server_default=func.now(), onupdate=func.now()))


class AssayDetail(SQLModel, table=True):
    __tablename__ = "assay_details"
    __table_args__ = {"schema": DB_SCHEMA}

    assay_id: int = Field(foreign_key=f"{DB_SCHEMA}.assays.id", primary_key=True)
    property_id: int = Field(foreign_key=f"{DB_SCHEMA}.properties.id", primary_key=True)

    value_datetime: Optional[datetime] = Field(default=None, sa_column=Column(DateTime(timezone=True)))
    value_uuid: Optional[uuid.UUID] = Field(default_factory=uuid.uuid4)
    value_num: Optional[float] = Field(default=None)
    value_string: Optional[str] = Field(default=None)

    assay: "Assay" = Relationship(back_populates="assay_details")
    property: "Property" = Relationship(back_populates="assay_details")


class AssayResponse(AssayResponseBase):
    properties: Optional[List["PropertyWithValue"]] = []
    property_requirements: List["AssayProperty"] = []


class Assay(AssayResponseBase, table=True):
    __tablename__ = "assays"
    __table_args__ = {"schema": DB_SCHEMA}

    created_by: uuid.UUID = Field(foreign_key=f"{DB_SCHEMA}.users.id", nullable=False, default_factory=uuid.uuid4)
    updated_by: uuid.UUID = Field(foreign_key=f"{DB_SCHEMA}.users.id", nullable=False, default_factory=uuid.uuid4)

    properties: List["Property"] = Relationship(
        back_populates="assays", link_model=AssayDetail, sa_relationship_kwargs={"viewonly": True}
    )

    assay_runs: List["AssayRun"] = Relationship(back_populates="assay")
    assay_details: List["AssayDetail"] = Relationship(back_populates="assay")
    property_requirements: List["AssayProperty"] = Relationship(back_populates="assay")


class AssayRunBase(SQLModel):
    name: str = Field(nullable=False)
    description: Optional[str] = Field(default=None)
    assay_id: int = Field(foreign_key=f"{DB_SCHEMA}.assays.id", nullable=False)


class AssayRunResponseBase(AssayRunBase):
    id: int = Field(primary_key=True, index=True)
    created_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now()))


class AssayRunResponse(AssayRunResponseBase):
    assay: Optional["Assay"] = None
    assay_run_details: List["AssayRunDetail"] = []
    properties: List["Property"] = []


class AssayRunCreate(AssayRunBase):
    property_ids: List[int] = []  # List of property IDs to associate with this assay


class AssayRunDetail(SQLModel, table=True):
    __tablename__ = "assay_run_details"
    __table_args__ = {"schema": DB_SCHEMA}

    assay_run_id: int = Field(foreign_key=f"{DB_SCHEMA}.assay_runs.id", primary_key=True)
    property_id: int = Field(foreign_key=f"{DB_SCHEMA}.properties.id", primary_key=True)

    value_datetime: Optional[datetime] = Field(default=None, sa_column=Column(DateTime(timezone=True)))
    value_uuid: Optional[uuid.UUID] = Field(default_factory=uuid.uuid4)
    value_num: Optional[float] = Field(default=None)
    value_string: Optional[str] = Field(default=None)

    assay_run: "AssayRun" = Relationship(back_populates="assay_run_details")
    property: Optional["Property"] = Relationship()


class AssayRun(AssayRunResponseBase, table=True):
    __tablename__ = "assay_runs"
    __table_args__ = {"schema": DB_SCHEMA}

    updated_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now()))
    created_by: uuid.UUID = Field(foreign_key=f"{DB_SCHEMA}.users.id", nullable=False, default_factory=uuid.uuid4)
    updated_by: uuid.UUID = Field(foreign_key=f"{DB_SCHEMA}.users.id", nullable=False, default_factory=uuid.uuid4)

    # Relationships - use assay_properties via assay to get list of expected properties
    # No direct properties relationship as assay_properties table no longer exists
    assay: "Assay" = Relationship(back_populates="assay_runs")
    assay_results: List["AssayResult"] = Relationship(back_populates="assay_run")
    assay_run_details: List["AssayRunDetail"] = Relationship(back_populates="assay_run")

    properties: List["Property"] = Relationship(
        back_populates="assay_runs",
        link_model=AssayRunDetail,
        sa_relationship_kwargs={"lazy": "joined", "viewonly": True},
    )


class AssayResultDetail(SQLModel, table=True):
    __tablename__ = "assay_result_details"
    __table_args__ = {"schema": DB_SCHEMA}

    assay_result_id: int = Field(foreign_key=f"{DB_SCHEMA}.assay_results.id", primary_key=True)
    property_id: int = Field(foreign_key=f"{DB_SCHEMA}.properties.id", primary_key=True)

    value_qualifier: Optional[int] = Field(default=0, nullable=False)  # 0 for "=", 1 for "<", 2 for ">"
    value_num: Optional[float] = Field(default=None)
    value_string: Optional[str] = Field(default=None)
    value_bool: Optional[bool] = Field(default=None)

    assay_result: "AssayResult" = Relationship(back_populates="assay_result_details")
    property: Optional["Property"] = Relationship(back_populates="assay_result_details")


class Property(PropertyResponse, table=True):
    __tablename__ = "properties"
    __table_args__ = (
        CheckConstraint(
            "value_type IN ('int', 'double', 'bool', 'datetime', 'string', 'uuid')", name="properties_value_type_check"
        ),
        CheckConstraint(
            "property_class IN ('DECLARED', 'CALCULATED', 'MEASURED', 'PREDICTED')",
            name="properties_property_class_check",
        ),
        CheckConstraint(
            "entity_type IN ('BATCH', 'COMPOUND', 'ASSAY', 'ASSAY_TYPES', 'ASSAY_RESULT', 'SYSTEM')",
            name="properties_entity_type_check",
        ),
        {"schema": DB_SCHEMA},
    )

    updated_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now()))
    created_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    updated_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)

    semantic_type: "SemanticType" = Relationship(back_populates="properties")
    min: Optional[float] = Field(default=None)
    max: Optional[float] = Field(default=None)
    choices: Optional[str] = Field(default=None)
    validators: Optional[str] = Field(default=None)

    batch_details: List["BatchDetail"] = Relationship(back_populates="property")
    compound_details: List["CompoundDetail"] = Relationship(back_populates="property")
    assay_details: List["AssayDetail"] = Relationship(back_populates="property")
    assay_result_details: List["AssayResultDetail"] = Relationship(back_populates="property")

    assays: List["Assay"] = Relationship(link_model=AssayDetail, sa_relationship_kwargs={"viewonly": True})
    assay_runs: List["AssayRun"] = Relationship(
        back_populates="properties", link_model=AssayRunDetail, sa_relationship_kwargs={"viewonly": True}
    )
    assay_results: List["AssayResult"] = Relationship(
        back_populates="properties", link_model=AssayResultDetail, sa_relationship_kwargs={"viewonly": True}
    )
    compounds: List["Compound"] = Relationship(
        back_populates="properties", link_model=CompoundDetail, sa_relationship_kwargs={"viewonly": True}
    )
    batches: List["Batch"] = Relationship(
        back_populates="properties", link_model=BatchDetail, sa_relationship_kwargs={"viewonly": True}
    )


class Validator(SQLModel, table=True):
    __tablename__ = "validators"
    __table_args__ = (
        CheckConstraint(
            "entity_type IN ('BATCH', 'COMPOUND', 'ASSAY', 'ASSAY_TYPES', 'ASSAY_RESULT', 'SYSTEM')",
            name="validators_entity_type_check",
        ),
        {"schema": DB_SCHEMA},
    )  # since table is in schema moltrack
    id: int = Field(primary_key=True, index=True)
    updated_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now()))
    created_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    updated_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    name: str = Field(unique=True, nullable=False)
    description: Optional[str] = Field(default=None)
    entity_type: enums.EntityType = Field(sa_column=Column(Enum(enums.EntityType), nullable=False))
    expression: str = Field(nullable=False)


class AssayResultBase(SQLModel):
    batch_id: int = Field(foreign_key=f"{DB_SCHEMA}.batches.id", nullable=False)
    assay_run_id: int = Field(foreign_key=f"{DB_SCHEMA}.assay_runs.id", nullable=False)
    updated_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now()))
    created_by: uuid.UUID = Field(foreign_key=f"{DB_SCHEMA}.users.id", nullable=False, default_factory=uuid.uuid4)
    updated_by: uuid.UUID = Field(foreign_key=f"{DB_SCHEMA}.users.id", nullable=False, default_factory=uuid.uuid4)


class AssayResultResponseBase(AssayResultBase):
    id: int = Field(primary_key=True, index=True)


# Extended response model with backward compatibility field
class AssayResultResponse(AssayResultResponseBase):
    # Add a computed field for backward compatibility
    result_value: Optional[Union[float, str, bool]] = None
    properties: List["PropertyWithValue"] = []

    @field_validator("result_value")
    def compute_result_value(cls, v, values):
        """Compute result_value from the appropriate typed value field"""
        if "value_num" in values and values["value_num"] is not None:
            return values["value_num"]
        elif "value_string" in values and values["value_string"] is not None:
            return values["value_string"]
        elif "value_bool" in values and values["value_bool"] is not None:
            return values["value_bool"]
        return None


class AssayResult(AssayResultResponseBase, table=True):
    __tablename__ = "assay_results"
    __table_args__ = {"schema": DB_SCHEMA}

    batch: "Batch" = Relationship(back_populates="assay_results")
    assay_run: "AssayRun" = Relationship(back_populates="assay_results")
    assay_result_details: List["AssayResultDetail"] = Relationship(back_populates="assay_result")

    properties: List["Property"] = Relationship(
        back_populates="assay_results",
        link_model=AssayResultDetail,
        sa_relationship_kwargs={"lazy": "joined", "viewonly": True},
    )


class BatchAssayResultsCreate(SQLModel):
    assay_run_id: int
    batch_id: int
    measurements: Dict[str, Union[float, str, bool, Dict[str, Any]]]


class BatchAssayResultsResponse(SQLModel):
    assay_run_id: int
    batch_id: int
    assay_name: str
    measurements: Dict[str, Union[float, str, bool, Dict[str, Any]]]


class AdditionFields(SQLModel):
    name: Optional[str] = Field(default=None, unique=True)
    description: Optional[str] = None
    code: Optional[str] = None
    smiles: Optional[str] = None
    role: Optional[enums.AdditionsRole] = Field(default=None, sa_column=Column(Enum(enums.AdditionsRole)))
    molfile: Optional[str] = None
    formula: Optional[str] = None
    # Issue related to alias not working properly in SQLModel: https://github.com/fastapi/sqlmodel/issues/374
    molecular_weight: Optional[float] = Field(
        default=None,
        schema_extra={"validation_alias": "molecular weight"},
    )

    model_config = ConfigDict(populate_by_name=True)


class AdditionBase(AdditionFields):
    name: str = Field(nullable=False, unique=True)
    role: enums.AdditionsRole = Field(sa_column=Column(Enum(enums.AdditionsRole)))


class AdditionUpdate(AdditionFields):
    pass


class Addition(AdditionBase, table=True):
    __tablename__ = "additions"
    __table_args__ = (
        CheckConstraint("role IN ('SALT', 'SOLVATE')", name="additions_role_check"),
        {"schema": DB_SCHEMA},
    )

    id: int = Field(primary_key=True, index=True)
    created_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now(), nullable=False))
    updated_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now(), onupdate=func.now(), nullable=False)
    )
    created_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    updated_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    is_active: bool = Field(default=True)
    is_archived: bool = Field(default=False)
    deleted_at: Optional[datetime] = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now()))
    deleted_by: Optional[uuid.UUID] = Field(default=None)


class BatchAdditionBase(SQLModel):
    batch_id: int = Field(foreign_key="moltrack.batches.id", nullable=False, unique=True)
    addition_id: int = Field(foreign_key="moltrack.additions.id", nullable=False, unique=True)
    addition_equivalent: float = Field(default=1)


class BatchAddition(BatchAdditionBase, table=True):
    __tablename__ = "batch_additions"
    __table_args__ = {"schema": DB_SCHEMA}

    id: int = Field(primary_key=True, index=True)
    created_at: datetime = Field(sa_column=Column(DateTime(timezone=True), server_default=func.now(), nullable=False))
    updated_at: datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now(), onupdate=func.now(), nullable=False)
    )
    created_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)
    updated_by: uuid.UUID = Field(nullable=False, default_factory=uuid.uuid4)

    batch: "Batch" = Relationship(back_populates="batch_additions")


class SchemaPayload(SQLModel):
    properties: List["PropertyInput"] = Field(default_factory=list)
    synonym_types: List["SynonymTypeBase"] = Field(default_factory=list)


class AssayPropertiesPayload(SQLModel):
    assay_type_details_properties: List["PropertyBase"] = Field(default_factory=list)
    assay_details_properties: List["PropertyBase"] = Field(default_factory=list)
    assay_type_properties: List["PropertyBase"] = Field(default_factory=list)


class AdditionsPayload(SQLModel):
    additions: List["AdditionBase"] = Field(default_factory=list)


class SchemaBatchResponse(SQLModel):
    properties: Optional[List["PropertyBase"]] = Field(default_factory=list)
    synonym_types: Optional[List["SynonymTypeBase"]] = Field(default_factory=list)
    additions: List["AdditionBase"] = Field(default_factory=list)


class AssayTypeCreateBase(SQLModel):
    name: str
    description: Optional[str] = None
    details: Dict[str, Any]


class AssayTypeCreate(SQLModel):
    assay_type: AssayTypeCreateBase


class AssayResultProperty(SQLModel):
    name: str
    required: bool


class AssayCreateBase(AssayBase):
    assay_result_properties: List[AssayResultProperty]
    extra_fields: Dict[str, Any] = Field(default_factory=dict)

    model_config = {"extra": "allow"}

    @model_validator(mode="before")
    def collect_extra_fields(cls, values):
        known_keys = {"name", "assay_result_properties"}
        extra = {k: v for k, v in values.items() if k not in known_keys}
        values["extra_fields"] = extra
        return values


class UpdateCheckResult(NamedTuple):
    action: str
    update_data: Optional[Dict[str, Any]] = None


def validate_field(v: str) -> str:
    # Basic field format validation
    if not v or not isinstance(v, str):
        raise ValueError("Field must be a non-empty string")

    # Check for valid field format (table.field or table.details.property)
    parts = v.split(".")
    if len(parts) < 2 or len(parts) > 3 or (len(parts) == 3 and parts[1] != "details"):
        raise ValueError("Field must be in format 'table.field' or 'table.details.property'")

    valid_tables = get_args(Level)
    valid_tables = valid_tables + ("users",)
    if parts[0] not in valid_tables:
        allowed = ", ".join(valid_tables)
        raise ValueError(f"Invalid table: {parts[0]}. Must be one of {allowed}")

    return v


# Advanced Search Models - New Recursive Structure
class AtomicCondition(SQLModel):
    """Individual atomic search condition with field, operator, and value"""

    field: str  # e.g., "compounds.canonical_smiles", "compounds.details.chembl"
    operator: enums.CompareOp
    value: Any
    threshold: Optional[float] = None  # For similarity searches (e.g., molecular similarity)

    @field_validator("field")
    def validate_field_format(cls, v):
        return validate_field(v)

    @model_validator(mode="before")
    def validate_threshold(cls, values):
        # Validate threshold is provided for operators that require it
        if isinstance(values, AtomicCondition):
            operator = values.get("operator")
            threshold = values.get("threshold")
            if operator == enums.CompareOp.IS_SIMILAR:
                if threshold is None:
                    raise ValueError(f"Operator {operator.value} requires a threshold value")
            elif threshold is not None:
                raise ValueError(f"Threshold not supported for operator: {operator}")
        return values


class LogicalNode(SQLModel):
    """Logical node combining multiple filters with AND/OR operator"""

    operator: enums.LogicOp
    conditions: List["Filter"]

    @field_validator("conditions")
    def validate_conditions(cls, v):
        if not v or len(v) == 0:
            raise ValueError("LogicalNode must have at least one condition")
        if len(v) == 1:
            raise ValueError("LogicalNode with single condition is redundant, use AtomicCondition directly")
        return v


# Filter is a union type - can be either AtomicCondition or LogicalNode
Filter = Union[AtomicCondition, LogicalNode]


Level = Literal["compounds", "batches", "assay_results", "assay_runs", "assays"]


AggregationOp = Union[enums.AggregationNumericOp, enums.AggregationStringOp]


class Aggregation(SQLModel):
    field: str
    operation: AggregationOp

    @field_validator("field", mode="after")
    def validate_field_format(cls, v):
        return validate_field(v)


class BaseSearchRequest(SQLModel):
    output: List[str]  # Columns to return
    aggregations: Optional[List[Aggregation]] = Field(default_factory=list)
    filter: Optional[Filter] = None
    output_format: enums.SearchOutputFormat = enums.SearchOutputFormat.json
    limit: Optional[int] = None

    @field_validator("output")
    def validate_output(cls, v):
        if not v:
            raise ValueError("Output must specify at least one column")
        return v


class SearchRequest(BaseSearchRequest):
    """Main search request model with recursive filter structure"""

    level: Level


class SearchResponse(SQLModel):
    """Search response model"""

    status: str
    data: List[Dict[str, Any]]
    total_count: int
    level: str
    columns: List[str]


# Update forward references for recursive types
LogicalNode.model_rebuild()


class Token(NamedTuple):
    type: str
    value: Union[str, float, bool, None]


class VocabularyUpdateRequest(SQLModel):
    property_id: int
    allowed_values: list[str]
